import time
from threading import Thread

import numpy as np
import torch


class CloudComputeTrainer:
    def __init__(self, app, agent_profile):
        self.app = app
        self.agent_profile = agent_profile
        self.token = None
        self.running = False
        self.thread = None

    def start(self):
        self.token = self.app.start_runner(self.agent_profile.agent_id, self.agent_profile.agent.dump_parameters())
        if not self.token:
            return False

        self.running = True
        self.thread = Thread(target=self.run)
        self.thread.start()
        return True

    def run(self):
        while self.running:
            time.sleep(1)
            parameters = self.app.fetch_runner_parameters(self.token)
            if not parameters:
                self.running = False
                break

            self.agent_profile.agent.load_parameters(parameters)

    def experience(self, experience):
        payload = {}

        if 'agent_transition' in experience:
            observation, action, next_observation = experience['agent_transition']
            self.agent_profile.agent_experience.write(observation, action, next_observation)
            payload['agent_transition'] = {
                'observation': observation.tolist(),
                'action': observation.tolist(),
                'next_observation': next_observation.tolist(),
            }

        if 'user_transition' in experience:
            observation, action, next_observation = experience['user_transition']
            self.agent_profile.user_experience.write(observation, action, next_observation)
            payload['user_transition'] = {
                'observation': observation.tolist(),
                'action': observation.tolist(),
                'next_observation': next_observation.tolist(),
            }

        if 'rewards' in experience:
            segment1, segment2, preference = experience['rewards']
            self.agent_profile.reward_buffer.write(segment1, segment2, preference)
            payload['user_transition'] = {
                'segment1': segment1.tolist(),
                'segment2': segment2.tolist(),
                'preference': preference,
            }

        self.app.add_runner_experience(self.token, payload)

    def sample_segments(self, segment_size):
        segment1 = self.agent_profile.agent_experience.sample_segment(segment_size)
        segment2 = self.agent_profile.agent_experience.sample_segment(segment_size)
        return segment1, segment2

    def stop(self):
        self.running = False
        self.thread.join()
        self.app.stop_runner(self.token)


class LocalTrainer:
    def __init__(self, agent_profile):
        self.agent_profile = agent_profile
        self.thread = None
        self.running = False

    def start(self):
        self.running = True
        self.thread = Thread(target=self.run)
        self.thread.start()

    def run(self):
        starting = True
        experience_batch_size = self.agent_profile.config['EXPERIENCE_BATCH_SIZE']
        reward_batch_size = self.agent_profile.config['REWARD_BATCH_SIZE']

        def train(experience_buffer):
            o, a, n = experience_buffer.read(experience_batch_size)
            with torch.no_grad():
                r = self.agent_profile.reward_ensemble.predict(o, a).numpy()
            d = np.zeros(a.shape, dtype=np.float32)
            self.agent_profile.agent.train((o, a, r, n, d))

        while self.running:
            if self.agent_profile.agent_experience.size > 0:
                starting = False
                train(self.agent_profile.agent_experience)

            if self.agent_profile.user_experience.size > 0:
                starting = False
                train(self.agent_profile.user_experience)

            if self.agent_profile.reward_buffer.size > 0:
                starting = False
                s1, s2, p = self.agent_profile.reward_buffer.read(reward_batch_size)
                self.agent_profile.reward_ensemble.train(s1, s2, p)

            if starting:
                time.sleep(1)

    def experience(self, experience):
        if 'agent_transition' in experience:
            observation, action, next_observation = experience['agent_transition']
            self.agent_profile.agent_experience.write(observation, action, next_observation)
        if 'user_transition' in experience:
            observation, action, next_observation = experience['user_transition']
            self.agent_profile.user_experience.write(observation, action, next_observation)
        if 'rewards' in experience:
            segment1, segment2, preference = experience['rewards']
            self.agent_profile.reward_buffer.write(segment1, segment2, preference)

    def sample_segments(self, segment_size):
        segment1 = self.agent_profile.agent_experience.sample_segment(segment_size)
        segment2 = self.agent_profile.agent_experience.sample_segment(segment_size)
        return segment1, segment2

    def stop(self):
        self.running = False
        self.thread.join()



import json
import os
import subprocess
from datetime import datetime
from threading import Lock

import requests

from reinforcebotagent.agent import Agent
from reinforcebotagent import reward
from reinforcebotagent.replay_buffer import ExperienceReplayBuffer, RewardReplayBuffer


class AgentProfile:
    def __init__(self, app_config):
        self.app_config = app_config
        self.initialised = False
        self.loading_lock = Lock()
        self.agent = None
        self.action_mapping = None
        self.user_experience = None
        self.agent_experience = None
        self.reward_ensemble = None
        self.reward_buffer = None
        self.author = None
        self.agent_id = None
        self.name = None
        self.description = None

    def create_buffers(self):
        observation_space = self.app_config['OBSERVATION_SPACE']
        segment_size = self.app_config['SEGMENT_SIZE']
        ensemble_size = self.app_config['ENSEMBLE_SIZE']
        experience_buffer_size = self.app_config['EXPERIENCE_BUFFER_SIZE']
        reward_buffer_size = self.app_config['REWARD_BUFFER_SIZE']

        if self.user_experience is None:
            self.user_experience = ExperienceReplayBuffer(observation_space, experience_buffer_size)

        self.agent_experience = ExperienceReplayBuffer(observation_space, experience_buffer_size)
        self.reward_ensemble = reward.Ensemble(observation_space, len(self.action_mapping), ensemble_size)
        self.reward_buffer = RewardReplayBuffer(observation_space, segment_size, reward_buffer_size)

    def load_initial_user_experience(self, action_mapping, user_experience):
        if self.initialised:
            raise ValueError('Agents cannot redefine their action space, a new agent must be created instead')

        self.loading_lock.acquire()
        self.initialised = True
        self.action_mapping = action_mapping
        self.user_experience = user_experience
        self.create_buffers()
        self.agent = Agent(self.app_config['OBSERVATION_SPACE'], len(action_mapping))
        self.loading_lock.release()

    def agent_path(self, create=True):
        author_path = os.path.join(self.app_config['AGENTS_PATH'],
                                   '_' if self.author is None else self.author['username'])
        path = os.path.join(author_path, self.name)
        if create:
            os.makedirs(path, exist_ok=True)
        return path

    def save(self, buffers=False):
        path = self.agent_path()

        if self.author is not None:
            with open(os.path.join(path, '..', 'settings.json'), 'w') as settings_file:
                json.dump({'id': self.author['id']}, settings_file, indent=2)

        with open(os.path.join(path, 'settings.json'), 'w') as settings_file:
            json.dump({
                'initialised': self.initialised,
                'action_mapping': {k: list(v) for k, v in self.action_mapping.items()} if self.action_mapping else None,
                'author': self.author,
                'agent_id': self.agent_id,
                'name': self.name,
                'description': self.description,
            }, settings_file, indent=2)

        if self.initialised:
            self.agent.save(os.path.join(path, 'parameters'))
            if buffers:
                buffers_path = os.path.join(path, 'buffers')
                os.makedirs(buffers_path, exist_ok=True)
                self.agent_experience.save(os.path.join(buffers_path, 'agent_experience'))
                self.user_experience.save(os.path.join(buffers_path, 'user_experience'))
                self.reward_buffer.save(os.path.join(buffers_path, 'rewards'))

    def backup(self):
        self.save(buffers=False)
        backups_path = os.path.join(self.agent_path(), 'backups')
        os.makedirs(backups_path, exist_ok=True)
        path = os.path.join(backups_path, f'{self.name}-{datetime.now().isoformat()}.tar.gz')
        tar_cmd = ['tar', '-czf', str(path), 'settings.json']
        if os.path.exists(os.path.join(self.agent_path(), 'parameters')):
            tar_cmd.append('parameters')
        subprocess.run(tar_cmd, cwd=self.agent_path())
        return path

    @staticmethod
    def load(app_config, name, author=None):
        author_username = author['username'] if author else '_'
        agent_path = os.path.join(app_config['AGENTS_PATH'], author_username, name)

        with open(os.path.join(agent_path, 'settings.json'), 'r') as settings_file:
            settings = json.load(settings_file)

        profile = AgentProfile(app_config)
        profile.initialised = settings['initialised']
        profile.author = settings['author']
        profile.agent_id = settings['agent_id']
        profile.name = settings['name']
        profile.description = settings['description']

        if profile.initialised:
            profile.action_mapping = {int(k): set(v) for k, v in settings['action_mapping'].items()}
            profile.agent = Agent.load(os.path.join(agent_path, 'parameters'))
            buffers_path = os.path.join(agent_path, 'buffers')
            if os.path.exists(buffers_path):
                profile.agent_experience = ExperienceReplayBuffer.load(os.path.join(buffers_path, 'agent_experience'))
                profile.user_experience = ExperienceReplayBuffer.load(os.path.join(buffers_path, 'user_experience'))
                profile.reward_buffer = RewardReplayBuffer.load(os.path.join(buffers_path, 'reward_buffer'))
            else:
                profile.create_buffers()
        return profile

    @staticmethod
    def download(app_config, app, agent_id):
        agent = app.authorised_fetch(lambda h: requests.get(app_config['API_URL'] + f'agents/{agent_id}/', headers=h))
        if agent is None:
            return None

        agent = agent.json()
        author = requests.get(app_config['API_URL'] + f'users/{agent["author"]}/').json()

        author_path = os.path.join(app_config['AGENTS_PATH'], author['username'])
        agent_path = os.path.join(author_path, agent['name'])
        tarfile = os.path.join(app_config['CACHE_PATH'], agent['name'] + '.tar.gz')

        os.makedirs(agent_path, exist_ok=True)
        with open(os.path.join(author_path, 'settings.json'), 'w') as settings_file:
            json.dump({'id': author['id']}, settings_file, indent=2)

        subprocess.run(['wget', agent['parameters'], '-O', tarfile])
        subprocess.run(['tar', '-xf', tarfile, '-C', str(agent_path)])

        with open(os.path.join(agent_path, 'settings.json'), 'r') as settings_file:
            agent_settings = json.load(settings_file)

        with open(os.path.join(agent_path, 'settings.json'), 'w') as settings_file:
            json.dump({
                **agent_settings,
                'author': author,
                'agent_id': agent_id,
                'name': agent['name'],
                'description': agent['description'],
            }, settings_file, indent=2)

        return AgentProfile.load(app_config, agent['name'], author)



import json
import os

import numpy as np


class RewardReplayBuffer:
    def __init__(self, observation_space, segment_size, max_size):
        self.o1 = np.empty((max_size, segment_size, *observation_space), dtype=np.float32)
        self.a1 = np.empty((max_size, segment_size), dtype=np.int8)
        self.o2 = np.empty((max_size, segment_size, *observation_space), dtype=np.float32)
        self.a2 = np.empty((max_size, segment_size), dtype=np.int8)
        self.p = np.empty((max_size,), dtype=np.float32)
        self.index = 0
        self.size = 0
        self.max_size = max_size
        self.observation_space = observation_space
        self.segment_size = segment_size

    def write(self, s1, s2, p):
        o1, a1 = s1
        o2, a2 = s2
        self.o1[self.index] = o1
        self.a1[self.index] = a1
        self.o2[self.index] = o2
        self.a2[self.index] = a2
        self.p[self.index] = p
        self.index = (self.index + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def read(self, batch_size):
        indices = np.random.randint(0, self.size, size=batch_size)
        return (
            (self.o1[indices], self.a1[indices]),
            (self.o2[indices], self.a2[indices]),
            self.p[indices],
        )

    def save(self, path):
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, 'settings.json'), 'w') as settings_file:
            json.dump({
                'index': self.index,
                'size': self.size,
                'max_size': self.max_size,
                'observation_space': self.observation_space,
                'segment_size': self.segment_size,
            }, settings_file, indent=2)

        np.save(os.path.join(path, 'segment1_observations.npy'), self.o1)
        np.save(os.path.join(path, 'segment1_actions.npy'), self.a1)
        np.save(os.path.join(path, 'segment2_observations.npy'), self.o2)
        np.save(os.path.join(path, 'segment2_actions.npy'), self.a2)
        np.save(os.path.join(path, 'preferences.npy'), self.p)

    @staticmethod
    def load(path):
        with open(os.path.join(path, 'settings.json'), 'r') as settings_file:
            settings = json.load(settings_file)

        buffer = RewardReplayBuffer(settings['observation_space'], settings['segment_size'], settings['max_size'])
        buffer.index = settings['index']
        buffer.size = settings['size']
        buffer.o1 = np.load(os.path.join(path, 'segment1_observations.npy'))
        buffer.a1 = np.load(os.path.join(path, 'segment1_actions.npy'))
        buffer.o2 = np.load(os.path.join(path, 'segment2_observations.npy'))
        buffer.a2 = np.load(os.path.join(path, 'segment2_actions.npy'))
        buffer.p = np.load(os.path.join(path, 'preferences.npy'))
        return buffer


class ExperienceReplayBuffer:
    def __init__(self, observation_space, max_size):
        self.o = np.empty((max_size, *observation_space), dtype=np.float32)
        self.a = np.empty((max_size,), dtype=np.int8)
        self.n = np.empty((max_size, *observation_space), dtype=np.float32)
        self.index = 0
        self.size = 0
        self.max_size = max_size
        self.observation_space = observation_space

    def write(self, o, a, n):
        self.o[self.index] = o
        self.a[self.index] = a
        self.n[self.index] = n
        self.index = (self.index + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def read(self, batch_size):
        indices = np.random.randint(0, self.size, size=batch_size)
        return (
            self.o[indices],
            self.a[indices],
            self.n[indices],
        )

    def sample_segment(self, size):
        start = np.random.randint(0, self.size - size)
        stop = start + size
        o = self.o[start:stop]
        a = self.a[start:stop]
        return o, a

    def save(self, path):
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, 'settings.json'), 'w') as settings_file:
            json.dump({
                'index': self.index,
                'size': self.size,
                'max_size': self.max_size,
                'observation_space': self.observation_space,
            }, settings_file, indent=2)

        np.save(os.path.join(path, 'observations.npy'), self.o)
        np.save(os.path.join(path, 'actions.npy'), self.a)
        np.save(os.path.join(path, 'next_observations.npy'), self.n)

    @staticmethod
    def load(path):
        with open(os.path.join(path, 'settings.json'), 'r') as settings_file:
            settings = json.load(settings_file)

        buffer = ExperienceReplayBuffer(settings['observation_space'], settings['max_size'])
        buffer.index = settings['index']
        buffer.size = settings['size']
        buffer.o = np.load(os.path.join(path, 'observations.npy'))
        buffer.a = np.load(os.path.join(path, 'actions.npy'))
        buffer.n = np.load(os.path.join(path, 'next_observations.npy'))
        return buffer


class DynamicExperienceReplayBuffer(ExperienceReplayBuffer):
    def __init__(self, observation_space, max_size):
        super(DynamicExperienceReplayBuffer, self).__init__(observation_space, max_size)
        self.a = [''] * max_size
        self.action_space = {'', }

    def write(self, o, a, n):
        a = ','.join(map(str, a))
        super(DynamicExperienceReplayBuffer, self).write(o, a, n)
        self.action_space.add(a)

    def build(self):
        conversion = {action: idx for idx, action in enumerate(self.action_space)}
        buffer = ExperienceReplayBuffer(self.observation_space, self.max_size)
        buffer.o = self.o
        buffer.n = self.n
        buffer.a[:self.size] = np.array([conversion[a] for a in self.a[:self.size]], dtype=np.int8)
        action_mapping = {
            idx: set(map(int, action.split(','))) if action else set()
            for idx, action in enumerate(self.action_space)
        }
        return action_mapping, buffer



import json
import os

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn, optim

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Critic(nn.Module):
    def __init__(self, in_dim, out_dim):
        super(Critic, self).__init__()
        steps, width, height = in_dim
        ks = 5
        stride = 2
        self.conv1 = nn.Conv2d(steps, 16, kernel_size=ks, stride=stride)
        self.bn1 = nn.BatchNorm2d(16)
        self.conv2 = nn.Conv2d(16, 32, kernel_size=ks, stride=stride)
        self.bn2 = nn.BatchNorm2d(32)
        self.conv3 = nn.Conv2d(32, 32, kernel_size=ks, stride=stride)
        self.bn3 = nn.BatchNorm2d(32)

        def convolved_size(size, layers=3):
            return size if layers <= 0 else \
                convolved_size((size - ks) // stride + 1, layers - 1)

        linear_size = convolved_size(width) * convolved_size(height) * 32
        self.fc1 = nn.Linear(linear_size, out_dim)

    def forward(self, x):
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        return self.fc1(x.view(x.size(0), -1)).squeeze()


class Agent:
    def __init__(self, observation_space, action_space,
                 alpha=0.001, epsilon=0.1, epsilon_decay=1.0, gamma=0.99, tau=0.005):
        self.observation_space = observation_space
        self.action_space = action_space
        self.critic_target = Critic(observation_space, self.action_space).to(device)
        self.critic = Critic(observation_space, self.action_space).to(device)
        self.critic_criterion = nn.MSELoss()
        self.critic_optimiser = optim.Adam(self.critic.parameters(), lr=alpha)
        self.critic_target.load_state_dict(self.critic.state_dict())
        self.critic_target.eval()
        self.alpha = alpha
        self.epsilon = epsilon
        self.epsilon_decay = epsilon_decay
        self.gamma = gamma
        self.tau = tau

    def act(self, observation, epsilon=0.1):
        if epsilon > np.random.rand():
            return np.random.randint(self.action_space)

        q = self.critic(torch.from_numpy(observation).unsqueeze(0).float().to(device))
        return int(q.argmax())

    def train(self, experience):
        # Q(s,a)=Q(s,a)-alpha*(r+gamma*max_a(Q(st+1,a))-Q(s,a))
        o, a, r, n, d = experience
        with torch.no_grad():
            actions = torch.from_numpy(a).long().view(-1, 1).to(device)
            future_q = self.critic_target(torch.from_numpy(n).float().to(device))
            future_q = future_q.gather(1, actions)
            future_q[d] = 0
            targets = torch.from_numpy(r).to(device) + self.gamma * future_q

        self.critic_optimiser.zero_grad()
        outputs = self.critic(torch.from_numpy(o).float().to(device))
        targets = outputs.scatter(1, actions, targets)
        loss = self.critic_criterion(outputs, targets)
        loss.backward()
        for param in self.critic.parameters():
            param.grad.data.clamp_(-1, 1)
        self.critic_optimiser.step()

    def update_targets(self):
        self.critic_target.load_state_dict(self.critic.state_dict())
        self.critic_target.eval()

    def dump_parameters(self):
        return [p.detach().cpu().numpy().tolist() for p in self.critic.parameters()]

    def load_parameters(self, parameters):
        for param, target_param in zip(parameters, self.critic.parameters()):
            target_param.data.copy_(torch.from_numpy(np.array(param)))
        self.critic.eval()
        self.update_targets()

    def save(self, path):
        os.makedirs(path, exist_ok=True)

        with open(os.path.join(path, 'settings.json'), 'w') as settings_file:
            json.dump({
                'observation_space': self.observation_space,
                'action_space': self.action_space,
                'alpha': self.alpha,
                'epsilon': self.epsilon,
                'epsilon_decay': self.epsilon_decay,
                'gamma': self.gamma,
                'tau': self.tau,
            }, settings_file, indent=2)

        torch.save({
            'critic_state_dict': self.critic.state_dict(),
            'critic_target_state_dict': self.critic_target.state_dict(),
        }, os.path.join(path, 'agent.pt'))

    @staticmethod
    def load(path):
        with open(os.path.join(path, 'settings.json'), 'r') as settings_file:
            settings = json.load(settings_file)

        agent = Agent(settings['observation_space'], settings['action_space'],
                      settings['alpha'], settings['epsilon'],
                      settings['epsilon_decay'], settings['gamma'], settings['tau'])
        checkpoint = torch.load(os.path.join(path, 'agent.pt'))
        agent.critic.load_state_dict(checkpoint['critic_state_dict'])
        agent.critic_target.load_state_dict(checkpoint['critic_target_state_dict'])
        agent.critic.eval()
        agent.critic_target.eval()
        return agent



import torch
import torch.nn.functional as F
from torch import nn, optim


class Predictor(nn.Module):
    def __init__(self, observation_space, action_space):
        super(Predictor, self).__init__()
        steps, width, height = observation_space
        ks = 5
        stride = 2
        self.conv1 = nn.Conv2d(steps, 16, kernel_size=ks, stride=stride)
        self.bn1 = nn.BatchNorm2d(16)
        self.conv2 = nn.Conv2d(16, 32, kernel_size=ks, stride=stride)
        self.bn2 = nn.BatchNorm2d(32)
        self.conv3 = nn.Conv2d(32, 32, kernel_size=ks, stride=stride)
        self.bn3 = nn.BatchNorm2d(32)

        def convolved_size(size, layers=3):
            return size if layers <= 0 else \
                convolved_size((size - ks) // stride + 1, layers - 1)

        linear_size = convolved_size(width) * convolved_size(height) * 32
        self.fc1 = nn.Linear(linear_size + action_space, 1)
        self.action_space = action_space

    def forward(self, x):
        observation, action = x
        action_encoded = torch.zeros(action.shape[0], self.action_space)
        action_encoded[torch.arange(action.shape[0]), action.long()] = 1

        x = F.relu(self.bn1(self.conv1(observation)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = torch.cat((action_encoded, x.view(x.size(0), -1)), dim=1)
        return self.fc1(x).squeeze()


class Ensemble:
    def __init__(self, observation_space, action_space, size):
        self.predictors = [Predictor(observation_space, action_space) for _ in range(size)]

    def predict(self, observation, action):
        observation = torch.from_numpy(observation).float()
        action = torch.from_numpy(action)
        total_reward = 0
        for predictor in self.predictors:
            total_reward += predictor((observation, action))
        mean_reward = total_reward / len(self.predictors)
        return mean_reward

    def train(self, segment1, segment2, preference):
        for predictor in self.predictors:
            optimiser = optim.Adam(predictor.parameters())
            optimiser.zero_grad()
            loss = torch.zeros(1)

            for o1, a1, o2, a2, hp in zip(*segment1, *segment2, preference):
                hp = float(hp)
                s1 = predictor((torch.from_numpy(o1), torch.from_numpy(a1))).sum()
                s2 = predictor((torch.from_numpy(o2), torch.from_numpy(a2))).sum()
                p1 = torch.exp(s1) / (torch.exp(s1) + torch.exp(s2))  # prob. segment 1 > segment 2
                p2 = torch.exp(s2) / (torch.exp(s1) + torch.exp(s2))  # prob. segment 2 > segment 1
                loss += - (hp * p1 + (1 - hp) * p2)

            loss.backward()
            for param in predictor.parameters():
                param.grad.data.clamp_(-1, 1)
            optimiser.step()
